with Apex_Occurrence_Types;
with Apex_Occurrence_Utils;
with Apex_Stack_Trace_Types;
with Apex_Stack_Trace;
with Apex_Stack_Trace_Foundation;
with Apex_System_Constants;
with Apex_Address_To_String;
with System;

package body Apex_Stack_Trace_Example is

    Exception_Traceback_Started : Boolean := False;


    function Generate_Trace_String
                (X : Apex_Occurrence_Types.Exception_Occurrence)
                return String is

        Frame_Count : Natural := 1;

        Frm : Apex_Stack_Trace_Types.Frame;
        Frame : Apex_Stack_Trace_Types.Frame_Access :=
           Apex_Stack_Trace_Types.To_Frame_Access (Frm'Address);
        Unwind_Complete : Boolean;

    begin
        -- Note that while it's possible to write this routine as a recursive
        -- function which returns a variable length string by traversing
        -- the call stack, that can take up a lot of stack space.
        --
        -- Instead, we make one pass at the stack counting up how many frames
        -- we hit before the top of the stack is reached...  once we have
        -- that number, we know how big of a string we need.  We allocate a
        -- string whose size is based on that number and fill it in with
        -- the pc's garnered from traversing the stack a second time.
        --
        -- It's not the fastest or prettiest, but it works and conserves
        -- stack space.

        Apex_Stack_Trace_Foundation.Initialize (Frame, X);
        Unwind_Complete := False;
        while not Unwind_Complete loop
            Apex_Stack_Trace_Foundation.Previous_Frame
               (Frame, Unwind_Complete);
            Frame_Count := Frame_Count + 1;
        end loop;

        declare
            -- The "+ 1" accounts for the leading zero which is always present
            Characters_Per_Pc : constant :=
               Apex_System_Constants.Size_Of_Address / 4 + 1;
            Spacing : constant := 1; -- number of spaces between PC's

            Buffer : String
                        (1 .. (Frame_Count * (Characters_Per_Pc + Spacing))) :=
               (others => ' ');
            Index : Natural := 1;
        begin
            Apex_Stack_Trace_Foundation.Initialize (Frame, X);
            Unwind_Complete := False;
            while not Unwind_Complete loop
                Buffer (Index .. Index + Characters_Per_Pc - 1) :=
                   Apex_Address_To_String
                      (Apex_Stack_Trace_Foundation.Program_Counter (Frame));
                Apex_Stack_Trace_Foundation.Previous_Frame
                   (Frame, Unwind_Complete);
                Index := Index + Characters_Per_Pc + Spacing;
            end loop;
            return Buffer;
        end;

    end Generate_Trace_String;

    function Extended_Exception_Information
                (X : Apex_Occurrence_Types.Exception_Occurrence)
                return String is
        X_Acc : Apex_Occurrence_Types.Exception_Occurrence_Access :=
           Apex_Occurrence_Types.To_Exception_Occurrence_Access (X'Address);
    begin
        return "Exception Name: " &
                  Apex_Occurrence_Utils.Exception_Name (X_Acc) & Ascii.Lf &
                  "Exception Message: " &
                  Apex_Occurrence_Utils.Exception_Message (X_Acc) & Ascii.Lf &
                  "Exception Call Trace: " &
                  Generate_Trace_String (X) & Ascii.Lf;
    end Extended_Exception_Information;


    procedure Unhandled_Trace_Dump
                 (Occurrence : Apex_Occurrence_Types.
                                  Exception_Occurrence_Access;
                  Occurred_In_Main_Task : Boolean) is
    begin
        if Occurred_In_Main_Task then
            Apex_Stack_Trace.Trace_Dump (Occurrence => Occurrence,
                                         Trace_Label => "_in_main_task");
        else
            Apex_Stack_Trace.Trace_Dump (Occurrence => Occurrence,
                                         Trace_Label => "_not_in_main_task");
        end if;
    end Unhandled_Trace_Dump;

    procedure Exception_Stack_Trace_Enable is
    begin
        if not Exception_Traceback_Started then
            Exception_Traceback_Started := True;
            Apex_Stack_Trace.Exception_Trace_Enable;
            Apex_Stack_Trace_Foundation.Exception_Callout_Install
               (Initial_Raise => System.Null_Address,
                Subsequent_Reraise => System.Null_Address,
                Unhandled => Unhandled_Trace_Dump'Address);
            Apex_Occurrence_Utils.Exception_Information_Callout_Install
               (Extended_Exception_Information'Address);
        end if;
    end Exception_Stack_Trace_Enable;

end Apex_Stack_Trace_Example;
